#include "Parsodus/driver.h"
#include "Parsodus/inputparser.h"
#include "Parsodus/config.h"

#include "g3log/g3log.hpp"
#include "g3log/logworker.hpp"

#include <fstream>
#include <iostream>
#include <sys/stat.h>

#if defined(_WIN32)
#include <io.h>
#define isatty(x) _isatty(x)
#define fileno(x) _fileno(x)
#include <direct.h>  
#else
#include <unistd.h>
#endif

namespace {
    /**
     * Filter only valid identifier chars: alphanumeric, and not starting with a digit
     */
    std::string clean(std::string in) {
        std::string s;
        for (char c : in) {
            if ((s.length() && std::isalnum(c)) || std::isalpha(c) || c == '_')
                s += c;
        }
        return s;
    }

    /**
     * G3Log sink
     */
    class Sink {
        public:
            Sink(std::unique_ptr<std::ostream>&& infoOut) : m_infoOut(std::move(infoOut)) {}
            void ReceiveLogMessage(g3::LogMessageMover logEntry) {
                auto level = logEntry.get().level();

                std::string prefix = "";
                std::string suffix = "";

                if (isatty(fileno(stdin))) {
                    prefix = "\033[38;5;202;1m";
                    suffix = "\033[0m";
                }

                if (level == WARNING.text) {
                    std::cerr << prefix << logEntry.get().message() << suffix;
                } else if (level == INFO.text || level == DEBUG.text) {
                    if (m_infoOut)
                        *m_infoOut << logEntry.get().message();
                }
            }
        private:
            std::unique_ptr<std::ostream> m_infoOut;
    };
}


namespace pds {
    
    Driver::Driver(std::unique_ptr<BackendManager> backends, std::istream& inputfile, std::string outputdir, std::string language, std::string parsername, bool debug):
        m_backends(std::move(backends)), m_inputfile(inputfile), m_outputdir(outputdir), m_language(language), m_parsername(clean(parsername)), m_debug(debug) {
        }

    Driver::~Driver(){}

    int Driver::run() { 
        struct stat sb;
        if(stat(m_outputdir.c_str(), &sb) != 0) {
            int status; 
#if defined(_WIN32)
            status = _mkdir(m_outputdir.c_str());
#else
            status = mkdir(m_outputdir.c_str(), 0755);
#endif
            if(status !=0 && errno != EEXIST){
                throw DriverException("The folder " + m_outputdir + " does not exist and we're unable to create it.");
            }
        }

        std::unique_ptr<g3::LogWorker> logworker{ g3::LogWorker::createLogWorker() };
        if (m_debug)
            logworker->addSink(std::make_unique<Sink>(std::make_unique<std::ofstream>(m_outputdir + "/debug.log")), &Sink::ReceiveLogMessage);
        else
            logworker->addSink(std::make_unique<Sink>(nullptr), &Sink::ReceiveLogMessage);
        g3::initializeLogging(logworker.get());

        if (!m_parsername.length()) throw DriverException("no valid parser name possible"); 
        Config config = InputParser::parseInput(m_inputfile);       
        Backend* back = m_backends->findBackend(m_language, config.parserType);
        if (!back) throw DriverException("Could not find a valid backend for language " + m_language + " and parser type: " + config.parserType);
		back->generateParser([this](std::string filename) -> std::unique_ptr<std::ostream> {
                return std::unique_ptr<std::ostream>(new std::ofstream(m_outputdir + "/" + filename));
				}, m_parsername, config);  

        return 0; 
    }

    DriverException::DriverException(std::string what): m_what(what) {} 
    const char* DriverException::what() const throw() {
        return m_what.c_str();
    }

}

